#include "core.h"
#include "alotalot.h"
#include <assert.h>

namespace nytl {
    struct EEFrame {
        const json::JSON& expr;
        LocalVarValue& result;
        LocalVarValue temp_ret;
        size_t chain_el = 0;

        EEFrame(const json::JSON &expr, LocalVarValue &result)
            : expr(expr),
              result(result) {
        }

        void descend(const json::JSON& what) {
            if (result.is_json) {
                const json::JSON& P = *result.JSON_subval;
                if (P.isArray() && what.isInteger()) {
                    const std::vector<json::JSON> arr_p = P.asArray();
                    int64_t ind_w = what.asInteger().get_int();
                    ASSERT(ind_w > 0 && ind_w < arr_p.size(), "Expression \"array[integer]\" caused out-of-bound situation");
                    result = LocalVarValue{true, "", &arr_p[ind_w]};
                } else if (P.isDictionary() && what.isString()) {
                    const std::map<std::string, json::JSON> dict_p = P.asDictionary();
                    const std::string& key_w = what.asString();
                    ASSERT(dict_p.count(key_w) == 1, "No such key exception");
                    result = LocalVarValue{true, "", &dict_p.at(key_w)};
                } else
                    THROW("Incorrect type of \"json[json]\" expression. Unallowed signature of [] operator");
            } else {
                ASSERT(what.isString(), "Expression \"element[X]\" allowed only if X is string (json object)");
                if (what.asString().empty())
                    return;
                if (!is_uname_dotted_sequence(what.asString()))
                    THROW("Incorrect X in \"element[X]\"");
                result.EL_name += ("." + what.asString());
            }
        }

        uptr<EEFrame> toMe(bool returned, const global_elem_set_t& global_elems,
                const std::vector<LocalVarValue>& local_vars) {
            if (returned) {
                ASSERT(temp_ret.is_json, "Expression \"X[ element ]\" is not allowed");
                assert(temp_ret.JSON_subval);
                descend(*(temp_ret.JSON_subval));
            } else {
                assert(expr.isDictionary());
                const json::JSON& val = expr["V"].g();
                if (val.isInteger()) {
                    size_t lv_ind = val.asInteger().get_int();
                    assert(lv_ind < local_vars.size());
                    result = local_vars[lv_ind];
                } else if (val.isString()) {
                    std::string cur_el_name_str = expr["V"].g().asString();
                    result = LocalVarValue{false, cur_el_name_str, NULL};
                } else
                    assert(false);
            }
            const std::vector<json::JSON>& chain = expr["C"].g().asArray();
            while (true) {
                if (chain_el >= chain.size())
                    return NULL;
                const json::JSON& t = chain[chain_el++];
                if (t.isDictionary())
                    return std::make_unique<EEFrame>(t, temp_ret);
                descend(t);
            }
        }
    };

    LocalVarValue rendering_core_execute_expression(const global_elem_set_t& global_elems,
        const std::vector<LocalVarValue>& local_vars, const json::JSON& expr) {
        bool returned = false;
        std::vector<uptr<EEFrame>> stack;
        LocalVarValue result;
        stack.push_back(std::make_unique<EEFrame>(expr, result));
        while (!stack.empty()) {
            EEFrame& cur = *stack.back();
            uptr<EEFrame> todo = cur.toMe(returned, global_elems, local_vars);
            returned = !(bool)todo;
            if (todo)
                stack.push_back(mv(todo));
            else
                stack.pop_back();
        }
        return result;
    }
}
